import copy
import datetime
import os
import numpy as np
import traceback

from random import randint
from typing import Tuple

from .refldata import ReflData
from .resolution import FWHM2sigma

WAVELENGTH = 2.35
WAVELENGTH_DISPERSION = 0.02

UNIT_MAP = {'slit1': 'mm',
            'slit2': 'mm',
            'Slit3': 'mm',
            '2Theta': 'deg',
            'Theta': 'deg'}

def tostr(b: bytes) -> str:
    return b.decode('ascii')

def _split_sections(data: str) -> Tuple[str, list[str]]:
    """Split the data into sections

    Args:
        data (str): file data

    Returns:
        str, list[str]: header and list of data sections
    """

    sections = data.split('#S')

    # first section is the header
    header = sections.pop(0)

    # add split string back in
    for i, section in enumerate(sections):
        sections[i] = '#S' + section

    return header, sections

def _parse_header(header: str) -> dict:
    """Parse the header

    Args:
        header (str): header string

    Returns:
        dict: header contents
    """

    res = {}

    for line in header.splitlines():
        line = line.rstrip()
        if len(line):
            code, data = line.split('#')[1].split(' ', 1)
            match code[0]:
                case 'F':
                    res.update({'name': data})
                case 'E':
                    res.update({'epoch': data})
                case 'D':
                    res.update({'timestamp': data})
                case 'O':
                    res.update({'header fields': res.get('header fields', []) + data.split()})

    return res

def _parse_section(section: str) -> dict:
    """Parse a section

    Args:
        section (str): section string

    Returns:
        dict: section contents
    """

    res = {}
    table_data = []

    for line in section.splitlines():
        line = line.rstrip()
        if not len(line):
            continue
        
        # header information
        if '#' in line:
            code, data = line.split('#')[1].split(' ', 1)
            match code[0]:
                case 'S':
                    sdata = data.split()
                    scan_number = int(sdata.pop(0))
                    intent = sdata.pop(0)
                    scan_time = float(sdata.pop(-1))
                    number_points = float(sdata.pop(-1))
                    scan_parameters = {}
                    while len(sdata):
                        name = sdata.pop(0)
                        start_value = float(sdata.pop(0))
                        end_value = float(sdata.pop(0))
                        scan_parameters.update({name: {'start': start_value, 'end': end_value}})

                    res.update({'scan_number': scan_number,
                                'intent': intent,
                                'scan_time': scan_time,
                                'expected_number_points': number_points,
                                'scan_parameters': scan_parameters})
                case 'D':
                    res.update({'timestamp': data})
                case 'P':
                    res.update({'header data': res.get('header data', []) + data.split()})
                case 'N':
                    res.update({'number data fields': float(data)})
                case 'L':
                    res.update({'data fields': data.split()})
        # data
        else:
            table_data.append([float(d) for d in line.split()])

    # update actual length of data
    res['actual_number_points'] = len(table_data)
    res['data'] = np.array(table_data, ndmin=2)

    return res

def load_entries(filename, file_obj=None, entries=None):
    """
    Load the entries from a GANS Spec data file.
    """

    if file_obj is None:
        with open(filename, 'r') as fid:
            data: str = fid.read()
    else:
        data: str = file_obj.read().decode('ascii')

    header, sections = _split_sections(data)

    if header[:2] != '#F':
        print(f'Invalid SPEC file {filename}')
        return []

    parsed_header = _parse_header(header)
    parsed_sections = [_parse_section(section) for section in sections if section[:2] == '#S']
    parsed_sections = [section for section in parsed_sections if section['actual_number_points'] > 0]
    ret_entries = [GANSRefl(section, parsed_header, filename) for section in parsed_sections]

    if entries is not None:
        ret_entries = [entry for entry in ret_entries if entry.entry in entries]

    return ret_entries

def entry_field(entry: dict, header: dict, entry_name: str):
    """Gets data from entry if it exists, otherwise from the header

    Args:
        entry (dict): parsed entry
        header (dict): parsed header
        entry_name (str): field name
    """

    if entry_name in entry['data fields']:
        data_index = entry['data fields'].index(entry_name)
        return entry['data'][:, data_index]
    if entry_name in header['header fields']:
        data_index = header['header fields'].index(entry_name)
        return float(entry['header data'][data_index]) * np.ones(entry['actual_number_points'])
    else:
        print(f'{entry_name} not found')

def _convert_slitrotation_to_aperture(rot: float, rot_zero: float = 0.0) -> float:
    """Converts a rotary slit rotation in degrees to a slit aperture in mm

    Args:
        rot (float): slit rotation (deg)
        rot_zero (float, optional): zero of slit rotation. Optional, default 0.0.

    Returns:
        float: slit aperture in mm
    """

    slit_aperture = 27.2024 * np.cos(np.radians(31.1036 + 44.9851 + rot - rot_zero)) - 6.54
                                         
    return slit_aperture

class GANSRefl(ReflData):

    def __init__(self, entry, header, filename):
        super(GANSRefl, self).__init__()
        self._set_metadata(entry, header, filename)
        self.load(entry, header)

    def _set_metadata(self, entry, header, filename):
        self.entry = '%s_%d' % (header['name'], entry['scan_number'])
        self.path = os.path.abspath(filename)
        self.name = header['name']

        self.filenumber = entry['scan_number']

        #self.date = iso8601.parse_date(entry['start_time'][0].decode('utf-8'))
        self.date = datetime.datetime.strptime(entry['timestamp'], '%a %b %d %H:%M:%S %Y')
        self.description = self.entry
        self.instrument = 'GANS'

        # Determine the number of points in the scan.
        self.points = len(entry['data'])

        self.monitor.deadtime = 0.0
        self.monitor.deadtime_error = 0.0

        self.monitor.time_step = 0.001  # assume 1 ms accuracy on reported clock
        # Monitor
        self.monitor.counts = entry_field(entry, header, 'Monitor')
        self.monitor.counts_variance = self.monitor.counts.copy()
        self.monitor.count_time = entry_field(entry, header, 'Seconds')
        self.monitor.roi_counts = entry_field(entry, header, 'Detector')
        self.monitor.roi_variance = self.monitor.roi_counts.copy()

        # Needed?
        self.sample.name = header['name']
        self.sample.description = self.description

        self.scan_value = []
        self.scan_units = []
        self.scan_label = []
        
        scanned_variables = entry['data fields'][:entry['data fields'].index('H')]
        for var in scanned_variables:
            self.scan_value.append(entry_field(entry, header, var))
            self.scan_units.append(UNIT_MAP.get(var, None))
            self.scan_label.append(var)

    def load(self, entry, header):
        n = entry['actual_number_points']
        twotheta = entry['scan_parameters'].get('tth', None)
        theta = entry['scan_parameters'].get('th', None)
        chi = entry['scan_parameters'].get('ch', None)
        slit1 = entry['scan_parameters'].get('s1', None)
        if twotheta is not None:
            starting_twotheta = twotheta['start']
            if theta is not None:
                starting_theta = theta['start']
                if np.isclose(starting_theta, starting_twotheta / 2.0, rtol=1e-4):
                    self.intent = 'specular'
                elif (starting_theta > starting_twotheta / 2.0):
                    self.intent = 'background+'
                elif (starting_theta < starting_twotheta / 2.0):
                    self.intent = 'background-'
            else:
                self.intent = 'rock detector'
        elif theta is not None:
            self.intent = 'rock sample'
        elif slit1 is not None:
            self.intent = 'intensity'
        elif chi is not None:
            self.intent = 'rock chi'
        else:
            self.intent = 'scan'

        # TODO: Polarizers

        # Monochromator
        self.monochromator.wavelength = WAVELENGTH
        self.monochromator.wavelength_resolution = FWHM2sigma(WAVELENGTH_DISPERSION*self.monochromator.wavelength)

        # Slits
        self.slit1.distance = -1911
        self.slit2.distance = -200
        self.slit3.distance = 282
        self.slit4.distance = 921
        for lbl, slit in zip(['slit1', 'slit2', 'Slit3'], [self.slit1, self.slit2, self.slit4]):
            slit.x = entry_field(entry, header, lbl)
            slit.x_target = copy.copy(slit.x)
            slit.y = 50
            slit.y_target = 50

        self.slit3.x = 12.7 * np.ones(entry['actual_number_points'])
        self.slit3.x_target = copy.copy(self.slit3.x)

        # Detector
        self.detector.wavelength = self.monochromator.wavelength
        self.detector.wavelength_resolution = self.monochromator.wavelength_resolution
        self.detector.deadtime = np.array([0.0])
        self.detector.deadtime_error = np.array([0.0])
        self.detector.distance = 1000.0
        self.detector.rotation = 0.0

        # Counts
        self.detector.counts = entry_field(entry, header, 'Detector')
        self.detector.counts_variance = self.detector.counts.copy()
        self.detector.dims = self.detector.counts.shape[1:]

        # Angles
        self.sample.angle_x = entry_field(entry, header, 'Theta')
        self.detector.angle_x = entry_field(entry, header, '2Theta')
        self.sample.angle_x_target = self.sample.angle_x.copy()
        self.detector.angle_x_target = self.detector.angle_x.copy()
        self.sample.angle_y = entry_field(entry, header, 'Chi')

def demo():
    import sys
    from .load import setup_fetch, fetch_uri
    from .scale import apply_norm
    from .steps import divergence
    if len(sys.argv) == 1:
        print("usage: python -m reflred.gans file...")
        sys.exit(1)
    setup_fetch()
    plotted_datasets = 0
    for uri in sys.argv[1:]:
        try:
            entries = fetch_uri(uri, loader=load_entries)
        except Exception as exc:
            print("Error while loading", uri, ':', str(exc))
            #traceback.print_exc(); raise
            continue

        # print the first entry
        #print(entries[0])

        # plot all the entries
        #pylab.figure()
        for entry in entries:
            entry = divergence(entry)
            apply_norm(entry, base='time')
            entry.plot()
            plotted_datasets += 1

    if plotted_datasets:
        import pylab
        pylab.legend()
        pylab.show()
    else:
        print("no data to plot")

if __name__ == "__main__":
    demo()



import copy
import numpy as np

from reductus.reflred.steps import subtract_background, divide_intensity, rescale, mask_points, join, divergence, normalize, abinitio_footprint, fit_footprint, correct_footprint, fit_background_field, subtract_background_field

from reductus.dataflow.automod import module
from reductus.reflred.gansref import _convert_slitrotation_to_aperture


@module
def load(filelist=None,
        intent='auto',
        Qz_basis='detector',
        sample_width=None,
        slit1_distance=None,
        slit1_aperture=None,
        slit1_rotary=True,
        slit2_distance=None,
        slit2_aperture=None,
        slit2_rotary=True,
        slit3_distance=None,
        slit3_aperture=None,
        slit3_rotary=False,
        slit4_distance=None,
        slit4_aperture=None,
        slit4_rotary=False        
        ):
    r"""
    Load a list of GANS Reflectivity files.

    *Qz_basis* uses one of the following values:

        **detector**
            ignores the sample angle and calculates Qz
            as $(4\pi/\lambda \sin(\theta_\text{detector}/2))$,
        **sample**
            ignores the detector angle and calculates Qz
            as $(4\pi/\lambda \sin(\theta_\text{sample}))$
        **actual**
            calculates Qx and Qz as (x,z)-components of
            $(\vec k_{\text{out}} - \vec k_\text{in})$ in sample coordinates,

    **Inputs**

    filelist (fileinfo[]): List of files to open.

    intent (opt:auto|specular|background+\|background-\|intensity|rock sample|rock detector|rock chi|slit align|scan)
    : Measurement intent (specular, background+, background-, slit, rock),
    auto or infer.  If intent is 'scan', then use the first scanned variable.

    Qz_basis (opt:detector|sample|actual)
    : How to calculate Qz from instrument angles.

    sample_width {Sample width (mm)} (float?)
    : Width of the sample along the beam direction in mm, used for
    calculating the effective resolution when the sample is smaller
    than the beam.  Leave blank to use value from data file.

    slit1_distance {override slit1 distance} (float?)
    : if specified, will override the value found in the file for
    the distance from the sample to slit 1

    slit1_aperture {override slit1 aperture} (float?)
    : if specified, will override the value found in the file for
    the opening of slit 1 in mm

    slit1_rotary {slit1 rotary?} (bool)
    : designates slit 1 as a rotary slit and converts its value from
    slit rotation to a slit aperture in mm

    slit2_distance {override slit2 distance} (float?)
    : if specified, will override the value found in the file for
    the distance from the sample to slit 2

    slit2_aperture {override slit2 aperture} (float?)
    : if specified, will override the value found in the file for
    the opening of slit 2 in mm

    slit2_rotary {slit2 rotary?} (bool)
    : designates slit 2 as a rotary slit and converts its value from
    slit rotation to a slit aperture in mm    

    slit3_distance {override slit3 distance} (float?)
    : if specified, will override the value found in the file for
    the distance from the sample to slit 3

    slit3_aperture {override slit3 aperture} (float?)
    : if specified, will override the value found in the file for
    the opening of slit 3 in mm

    slit3_rotary {slit3 rotary?} (bool)
    : designates slit 3 as a rotary slit and converts its value from
    slit rotation to a slit aperture in mm

    slit4_distance {override slit4 distance} (float?)
    : if specified, will override the value found in the file for
    the distance from the sample to slit 4

    slit4_aperture {override slit4 aperture} (float?)
    : if specified, will override the value found in the file for
    the opening of slit 4 in mm    

    slit4_rotary {slit4 rotary?} (bool)
    : designates slit 4 as a rotary slit and converts its value from
    slit rotation to a slit aperture in mm

    **Returns**

    output (refldata[]): All entries of all files in the list.

    | 2024-11-26 David Hoogerheide
    """
    from reductus.reflred.load import url_load_list
    from reductus.reflred import gansref
    
    # TODO: sample_width is ignored if datafile defines angular_divergence
    auto_divergence = True
    enforce_specular = True

    datasets = []
    for data in url_load_list(filelist, loader=gansref.load_entries):
        data.Qz_basis = Qz_basis
        if intent not in [None, 'auto']:
            data.intent = intent
        if data.intent == 'specular' and enforce_specular:
            if data.Qz_basis == 'detector':
                data.sample.angle_x = data.sample.angle_x_target = data.detector.angle_x / 2.0
            elif data.Qz_basis == 'sample':
                data.detector.angle_x = data.detector.angle_x_target = data.sample.angle_x * 2.0
        if data.intent.startswith('background'):
            if data.Qz_basis == 'detector':
                data.Qz_target = 4 * np.pi / data.monochromator.wavelength * np.sin(np.radians(data.detector.angle_x_target / 2.0))
            elif data.Qz_basis == 'sample':
                data.Qz_target = 4 * np.pi / data.monochromator.wavelength * np.sin(np.radians(data.sample.angle_x_target))
            else:
                data.Qz_target = data.Qz
      
        if slit1_distance is not None:
            data.slit1.distance = slit1_distance
        if slit1_aperture is not None:
            data.slit1.x = data.slit1.x_target = slit1_aperture
        if slit1_rotary:
            data.slit1.x = _convert_slitrotation_to_aperture(data.slit1.x)
            data.slit1.x_target = copy.copy(data.slit1.x)
        if slit2_distance is not None:
            data.slit2.distance = slit2_distance
        if slit2_aperture is not None:
            data.slit2.x = data.slit2.x_target = slit2_aperture
        if slit2_rotary:
            data.slit2.x = _convert_slitrotation_to_aperture(data.slit2.x)
            data.slit2.x_target = copy.copy(data.slit2.x)            
        if slit3_distance is not None:
            data.slit3.distance = slit3_distance
        if slit3_aperture is not None:
            data.slit3.x = data.slit3.x_target = slit3_aperture
        if slit3_rotary:
            data.slit3.x = _convert_slitrotation_to_aperture(data.slit3.x)
            data.slit3.x_target = copy.copy(data.slit3.x)
        if slit4_distance is not None:
            data.slit4.distance = slit4_distance
        if slit4_aperture is not None:
            data.slit4.x = data.slit4.x_target = slit4_aperture
        if slit4_rotary:
            data.slit4.x = _convert_slitrotation_to_aperture(data.slit4.x)
            data.slit4.x_target = copy.copy(data.slit4.x)                        
        if auto_divergence:
            data = divergence(data, sample_width)
        
        data = normalize(data, base='time')
        print("data loaded and normalized")
        datasets.append(data)

    return datasets

@module
def fit_rocking_curve(rock, A0=None, x00=None, sigma0=None, bkg0=None):
    """
    Fit a rocking curve to a Gaussian plus background model.

    The rocking curve datasets are fit using a Levenberg-Marquardt algorithm to a model
    involving four parameters:
     o A, the (possibly negative) Gaussian amplitude
     o x0, the center of the Gaussian
     o sigma, the width of the Gaussian
     o bkg, the background level of the Gaussian

    **Inputs**

    rock (refldata) : single rocking curve

    A0 {Amplitude guess} (float) : initial guess for amplitude

    x00 {Gaussian center guess} (float) : initial guess for center

    sigma0 {width guess} (float) : initial guess for width

    bkg0 {background guess} (float) : initial guess for background
    
    **Returns**

    fitparams (gans.fitters.gaussianparams) : fit parameters, errors, and chi-squared

    fit (refldata) : ReflData structure containing fit outputs (for plotting against
    inputs to inspect fit)

    2024-12-04 David P. Hoogerheide
    """

    from .alignfit import fit_gaussian_background

    bff, rock2 = fit_gaussian_background(rock, A0, sigma0, x00, bkg0)

    return bff, rock2


@module
def fit_slit_alignment(slit_align, m0=None, x00=None):
    """
    Fit a slit alignment curve to a linear model with x-intercept.

    The slit alignment datasets are fit using a Levenberg-Marquardt algorithm to a model
    involving four parameters:
     o m, the linear slope
     o x0, the x-intercept of the line

    **Inputs**

    slit_align (refldata) : slit alignment curve

    m0 {slope guess} (float) : initial guess for slope

    x00 {x-intercept guess} (float) : initial guess for x intercept
    
    **Returns**

    fitparams (gans.fitters.linearparams) : fit parameters, errors, and chi-squared

    fit (refldata) : ReflData structure containing fit outputs (for plotting against
    inputs to inspect fit)

    2024-12-04 David P. Hoogerheide
    """

    from .alignfit import fit_line_xintercept

    bff, slit_align2 = fit_line_xintercept(slit_align, m0, x00)

    return bff, slit_align2



r"""
Curve fitting operations for alignment
"""
from __future__ import division

import numpy as np
import scipy.optimize as so
from copy import copy

# BaseFitResult class returns results of background field fitting

class BaseFitResult:

    def get_metadata(self):
        return {}
    
    def get_plottable(self):
        return {"params": self.get_metadata(), "type": "params"}

class ErrorFitResult(BaseFitResult):

    def __init__(self, msg):
        self.msg = msg

    def get_metadata(self):
        return {'error message': self.msg}

# ===============

class GaussianBackgroundFitResult(BaseFitResult):

    def __init__(self, chisq, A, sigma, x0, bkg, dA, dsigma, dx0, dbkg):
        self.chisq = chisq
        self.A = A
        self.dA = dA
        self.sigma = sigma
        self.dsigma = dsigma
        self.x0 = x0
        self.dx0 = dx0
        self.bkg = bkg
        self.dbkg = dbkg

    def get_metadata(self):
        return {
            "fit_chi-squared": self.chisq,
            "amplitude": self.A,
            "amplitude error": self.dA,
            "sigma": self.sigma,
            "sigma error": self.dsigma,
            "FWHM": self.sigma * 2.355,
            "FWHM error": self.dsigma * 2.355,
            "center": self.x0,
            "center error": self.dx0,
            "background": self.bkg,
            "background error": self.dbkg
        }

def gaussian_background(x, A, sigma, x0, bkg):
    """
    Calculates Gaussian with background
    """

    with np.errstate(divide='ignore'):
        return A * np.exp(-(x - x0) ** 2 / (2 * sigma ** 2)) + bkg
    
def gaussian_background_jac(x, A, sigma, x0, bkg):

    with np.errstate(divide='ignore'):
        g = np.exp(-(x - x0) ** 2 / (2 * sigma ** 2))
        return np.array([g,
                         A * g * (x - x0) ** 2 / (sigma ** 3),
                         A * g * (x - x0) / (2 * sigma ** 2),
                         np.ones_like(g)])
    
def fit_gaussian_background(rock, A0, sigma0, x00, bkg0):
    v, dv = rock.v, rock.dv
    if rock.intent == 'rock sample':
        x = rock.sample.angle_x
    elif rock.intent == 'rock detector':
        x = rock.detector.angle_x
    elif rock.intent == 'rock chi':
        x = rock.sample.angle_y
    else:
        return ErrorFitResult(msg=f'intent must be either "rock sample" or "rock detector" or "rock chi", actually {rock.intent}'), None

    if A0 is None:
        A0 = max(v)
    
    if sigma0 is None:
        sigma0 = np.sqrt(np.trapezoid(v * x ** 2, x) / np.trapezoid(v, x) - np.trapezoid(v * x, x) ** 2 / np.trapezoid(v, x) ** 2)
    
    if x00 is None:
        x00 = x[np.where(v==max(v))[0][0]]
    
    if bkg0 is None:
        bkg0 = min(v)

    def minfunc(pars):
        res = gaussian_background(x, *pars)
        return (res - v)/dv

    def jacfunc(pars):
        J = gaussian_background_jac(x, *pars)
        return (J*np.tile(1/dv, (len(pars), 1))).T

    print(f'Starting gaussian fit with initial parameters A={A0}, sigma={sigma0}, x0={x00}, bkg={bkg0}')

    pout, pcov, _, msg, ier = so.leastsq(
            minfunc,
            np.array([A0, sigma0, x00, bkg0]),
            Dfun=jacfunc,
            full_output=True, ftol=1e-15, xtol=1e-15)
    
    if ier in [1, 2, 3, 4]:
        # fit successful
        lenpout = len(pout)

        perr, chisq = np.sqrt(np.diag(pcov)), np.sum(minfunc(pout)**2) / (len(x) - lenpout)

        # create fit result data

        rock2 = copy(rock)
        rock2.v = gaussian_background(x, *pout)
        rock2.dv = np.sqrt(np.array([j.T.dot(pcov.dot(j)) for j in jacfunc(pout)]))

        return GaussianBackgroundFitResult(chisq, *pout, *perr), rock2

    else:

        return ErrorFitResult(msg), None

# ===============

class LineXInterceptFitResult(BaseFitResult):

    def __init__(self, chisq, m, x0, dm, dx0):
        self.chisq = chisq
        self.m = m
        self.dm = dm
        self.x0 = x0
        self.dx0 = dx0

    def get_metadata(self):
        return {
            "fit_chi-squared": self.chisq,
            "slope": self.m,
            "slope error": self.dm,
            "xintercept": self.x0,
            "xintercept error": self.dx0
        }

def line_xintercept(x, m, x0):
    """
    Calculates line with x-intercept
    """

    return m * (x - x0)

def line_xintercept_jac(x, m, x0):

    return np.array([x - x0, -m * np.ones_like(x)])

def fit_line_xintercept(slit_align, m0, x00):
    v, dv = slit_align.v, slit_align.dv
    x = None
    if slit_align.intent in ['slit align', 'intensity']:
        for slit in [slit_align.slit1, slit_align.slit2, slit_align.slit3, slit_align.slit4]:
            if any(bool(d) for d in np.diff(v)!=0):
                x = slit.x
                break

        if x is None:
            return ErrorFitResult(msg=f'no slits appear to be scanned'), None
        
    else:
        return ErrorFitResult(msg=f'intent must be "slit align" or "intensity", actually {slit_align.intent}'), None

    if m0 is None:
        m0 = (v[-1] - v[0]) / (x[-1] - x[0])
    
    if x00 is None:
        x00 = x[np.where(abs(v) == min(abs(v)))[0][0]]
    
    def minfunc(pars):
        res = line_xintercept(x, *pars)
        return (res - v)/dv

    def jacfunc(pars):
        J = line_xintercept_jac(x, *pars)
        return (J*np.tile(1/dv, (len(pars), 1))).T

    print(f'Starting linear fit with initial parameters m={m0}, x0={x00}')

    pout, pcov, _, msg, ier = so.leastsq(
            minfunc,
            np.array([m0, x00]),
            Dfun=jacfunc,
            full_output=True, ftol=1e-15, xtol=1e-15)
    
    if ier in [1, 2, 3, 4]:
        # fit successful
        lenpout = len(pout)

        perr, chisq = np.sqrt(np.diag(pcov)), np.sum(minfunc(pout)**2) / (len(x) - lenpout)

        # create fit result data

        slit_align2 = copy(slit_align)
        slit_align2.v = line_xintercept(x, *pout)
        slit_align2.dv = np.sqrt(np.array([j.T.dot(pcov.dot(j)) for j in jacfunc(pout)]))

        return LineXInterceptFitResult(chisq, *pout, *perr), slit_align2

    else:

        return ErrorFitResult(msg), None
